//========================================================================
// tb_SoftFlat
//========================================================================
// A Verilog test bench for our flat Sorting unit

`default_nettype none
`timescale 1ps/1ps


`include `"sort/`DESIGN.v`"
`include "vc/trace.v"
`include "vc/TestRandDelaySource.v"
`include "vc/TestRandDelaySink.v"

//------------------------------------------------------------------------
// Testbench defines
//------------------------------------------------------------------------

localparam NUM_TESTS = 3;

localparam         MSG_SIZE = 8;
localparam  INPUT_TEST_SIZE = 4*MSG_SIZE;
localparam OUTPUT_TEST_SIZE = 4*MSG_SIZE;

//------------------------------------------------------------------------
// Top-level module
//------------------------------------------------------------------------

module top(  input logic clk, input logic linetrace );
  
  // DUT signals
  logic        reset;

  logic                        istream_val;
  logic [ INPUT_TEST_SIZE-1:0] istream_msg;

  logic                        ostream_val;
  logic [OUTPUT_TEST_SIZE-1:0] ostream_msg;

  // Source and sink messages

  logic [  INPUT_TEST_SIZE-1:0 ] src_msgs [ NUM_TESTS-1:0 ];
  logic [ OUTPUT_TEST_SIZE-1:0 ] snk_msgs [ NUM_TESTS-1:0 ];

  // Signals to indicate completion

  logic src_done;
  logic snk_done;

  //----------------------------------------------------------------------
  // Module instantiations
  //----------------------------------------------------------------------

  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  // Test source
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  vc_TestSource 
  #(
    .p_msg_nbits ( INPUT_TEST_SIZE ),
    .p_num_msgs  (       NUM_TESTS )
  ) src (
    .clk         (             clk ),
    .reset       (           reset ),

    .val         (     istream_val ),
    .rdy         (            1'b1 ), // Always rdy
    .msg         (     istream_msg ),

    .done        (        src_done )
  );

  assign src.m = src_msgs;
  
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  // DUT
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  
  tut4_verilog_sort_`DESIGN #( MSG_SIZE )dut
  (
    .clk         (       clk   ),
    .reset       (       reset ),

    // Input stream

    .in_val ( istream_val ),
    .in0    ( istream_msg[     MSG_SIZE-1:         0] ),
    .in1    ( istream_msg[ (2*MSG_SIZE)-1:  MSG_SIZE] ),
    .in2    ( istream_msg[ (3*MSG_SIZE)-1:2*MSG_SIZE] ),
    .in3    ( istream_msg[ (4*MSG_SIZE)-1:3*MSG_SIZE] ),

    // Output stream

    .out_val ( ostream_val ),
    .out0    ( ostream_msg[     MSG_SIZE-1:         0] ),
    .out1    ( ostream_msg[ (2*MSG_SIZE)-1:  MSG_SIZE] ),
    .out2    ( ostream_msg[ (3*MSG_SIZE)-1:2*MSG_SIZE] ),
    .out3    ( ostream_msg[ (4*MSG_SIZE)-1:3*MSG_SIZE] )
  );

  initial begin 
    while( 1 ) begin
      @( negedge clk );  
      if( linetrace ) dut.display_trace;
    end 
    $stop;
   end

  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  // Test sink
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  vc_TestSink
  #(
    .p_msg_nbits ( OUTPUT_TEST_SIZE ),
    .p_num_msgs  (        NUM_TESTS ),
    .p_sim_mode  (                1 )
  ) sink (
    .clk         (              clk ),
    .reset       (            reset ),

    .val         (      ostream_val ),
    /* verilator lint_off PINCONNECTEMPTY */
    .rdy         (                  ), // Unused
    /* verilator lint_on PINCONNECTEMPTY */
    .msg         (      ostream_msg ),

    .done        (         snk_done )
  );

  assign sink.m = snk_msgs;

  //----------------------------------------------------------------------
  // Task for adding test cases
  //----------------------------------------------------------------------

  task test_case(
    input logic [  INPUT_TEST_SIZE-1:0 ] src_msg,
    input logic [ OUTPUT_TEST_SIZE-1:0 ] snk_msg
  );
  begin
    integer idx = 0;

    // Add messages to test arrays
    src_msgs[ idx ] = src_msg;
    snk_msgs[ idx ] = snk_msg;

    idx = idx + 1;
  end
  endtask

  //----------------------------------------------------------------------
  // Test cases
  //----------------------------------------------------------------------
  // Don't forget to change NUM_TESTS above when adding new tests!

  initial begin

    // Test cases

    test_case( {  8'd2, 8'd3, 8'd1, 8'd4 }, {  8'd4, 8'd3, 8'd2, 8'd1 } );
    test_case( { 8'd12, 8'd2, 8'd9, 8'd7 }, { 8'd12, 8'd9, 8'd7, 8'd2 } );
    test_case( {  8'd1, 8'd1, 8'd1, 8'd1 }, {  8'd1, 8'd1, 8'd1, 8'd1 } );

  end

  //----------------------------------------------------------------------
  // Run the Test Bench
  //----------------------------------------------------------------------

  initial begin

    $display( "Starting tb_SortFlat..." );
    reset = 1;
    
    // Wait a bit, then de-assert reset on negedge
    #10 
    @( negedge clk );
    reset = 0;

    // Wait for the test to finish
    while( !snk_done ) @( negedge clk );

    // Check that the source is also done
    if( !src_done )
      $error( "[ FAILED ] Our sink received more messages than our source has!" );
    else
      $display( "The testbench has finished" );

    // Delay for a bit for a better waveform
    #100
    $finish;
  end

  //----------------------------------------------------------------------
  // Timeout
  //----------------------------------------------------------------------
  // This is to ensure that our test eventually finishes, even if the sink
  // isn't receiving messages

  initial begin
    for( integer i = 0; i < 1000000; i = i + 1 ) begin
      @( negedge clk );
    end

    $error( "TIMEOUT: Testbench didn't finish in time" );
    $finish;
  end

endmodule
